#define IM_ROUND_8(x) ((int)((x)+0.5))
#define IM_ROUND_double(x) (x)
#define IM_LIMIT_8(x) ((x) < 0 ? 0 : (x) > 255 ? 255 : (x))
#define IM_LIMIT_double(x) ((x) < 0.0 ? 0.0 : (x) > 1.0 ? 1.0 : (x))
#line 1 "combine.im"
/*
=head1 NAME

combine.im - combining channels into an image

=head1 SYNOPSIS

  out = i_combine(imgs, channels, count);

=head1 DESCRIPTION

Combines channels from the input images into an output image.

=over

=cut
*/

#include "imager.h"

i_img *
i_combine(i_img **imgs, const int *channels, int in_count) {
  i_img *out = NULL;
  int maxbits = 0;
  i_img *maximg = NULL;
  int i;
  i_img_dim width, height;
  i_img_dim x, y;

  i_clear_error();
  if (in_count <= 0) {
    i_push_error(0, "At least one image must be supplied");
    return NULL;
  }
  if (in_count > MAXCHANNELS) {
    i_push_errorf(0, "Maximum of %d channels, you supplied %d",
		  MAXCHANNELS, in_count);
    return NULL;
  }

  width = imgs[0]->xsize;
  height = imgs[0]->ysize;
  for (i = 0; i < in_count; ++i) {
    if (imgs[i]->bits > maxbits) {
      maximg = imgs[i];
      maxbits = maximg->bits;
    }
    if (imgs[i]->xsize < width)
      width = imgs[i]->xsize;
    if (imgs[i]->ysize < height)
      height = imgs[i]->ysize;
    if (channels[i] < 0) {
      i_push_error(0, "Channel numbers must be zero or positive");
      return NULL;
    }
    if (channels[i] >= imgs[i]->channels) {
      i_push_errorf(0, "Channel %d for image %d is too high (%d channels)",
		    channels[i], i, imgs[i]->channels);
      return NULL;
    }
  }

  out = i_sametype_chans(maximg, width, height, in_count);
  if (!out)
    return NULL;
#line 66 "combine.im"
  if (maxbits <= i_8_bits) {
#undef IM_EIGHT_BIT
#define IM_EIGHT_BIT 1
#undef IM_FILL_COMBINE
#define IM_FILL_COMBINE(fill) ((fill)->combine)
#undef IM_FILL_FILLER
#define IM_FILL_FILLER(fill) ((fill)->f_fill_with_color)
#line 67 "combine.im"
  i_sample_t *in_row = mymalloc(sizeof(i_sample_t) * width);
  i_color *out_row = mymalloc(sizeof(i_color) * width);

  for (y = 0; y < height; ++y) {
    for (i = 0; i < in_count; ++i) {
      i_gsamp(imgs[i], 0, width, y, in_row, channels + i, 1);
      for (x = 0; x < width; ++x)
	out_row[x].channel[i] = in_row[x];
    }
    i_plin(out, 0, width, y, out_row);
  }
  myfree(out_row);
  myfree(in_row);
  }
  else {
#undef IM_EIGHT_BIT
#undef IM_FILL_COMBINE
#define IM_FILL_COMBINE(fill) ((fill)->combinef)
#undef IM_FILL_FILLER
#define IM_FILL_FILLER(fill) ((fill)->f_fill_with_fcolor)
#line 67 "combine.im"
  i_fsample_t *in_row = mymalloc(sizeof(i_fsample_t) * width);
  i_fcolor *out_row = mymalloc(sizeof(i_fcolor) * width);

  for (y = 0; y < height; ++y) {
    for (i = 0; i < in_count; ++i) {
      i_gsampf(imgs[i], 0, width, y, in_row, channels + i, 1);
      for (x = 0; x < width; ++x)
	out_row[x].channel[i] = in_row[x];
    }
    i_plinf(out, 0, width, y, out_row);
  }
  myfree(out_row);
  myfree(in_row);
  }
#line 81 "combine.im"

  return out;
}
