#define IM_ROUND_8(x) ((int)((x)+0.5))
#define IM_ROUND_double(x) (x)
#define IM_LIMIT_8(x) ((x) < 0 ? 0 : (x) > 255 ? 255 : (x))
#define IM_LIMIT_double(x) ((x) < 0.0 ? 0.0 : (x) > 1.0 ? 1.0 : (x))
#line 1 "conv.im"
#define IMAGER_NO_CONTEXT
#include "imager.h"
#include "imageri.h"

/*
  General convolution for 2d decoupled filters
  end effects are acounted for by increasing
  scaling the result with the sum of used coefficients

  coeff: (double array) coefficients for filter
    len: length of filter.. number of coefficients
           note that this has to be an odd number
           (since the filter is even);
*/

int
i_conv(i_img *im, const double *coeff,int len) {
  i_img_dim xo, yo; /* output pixel co-ordinate */
  int c, ch, center;
  double pc;
  double res[MAXCHANNELS];
  i_img *timg;
  dIMCTXim(im);

  im_log((aIMCTX,1,"i_conv(im %p, coeff %p, len %d)\n",im,coeff,len));
  im_clear_error(aIMCTX);

  if (len < 1) {
    im_push_error(aIMCTX, 0, "there must be at least one coefficient");
    return 0;
  }
 
  center=(len-1)/2;

  pc = 0;
  for (c = 0; c < len; ++c)
    pc += coeff[c];

  if (pc == 0) {
    i_push_error(0, "sum of coefficients is zero");
    return 0;
  }

  timg = i_sametype(im, im->xsize, im->ysize);

#line 46 "conv.im"
  if (im->bits <= 8) {
#undef IM_EIGHT_BIT
#define IM_EIGHT_BIT 1
#undef IM_FILL_COMBINE
#define IM_FILL_COMBINE(fill) ((fill)->combine)
#undef IM_FILL_FILLER
#define IM_FILL_FILLER(fill) ((fill)->f_fill_with_color)
#line 47 "conv.im"
  i_color rcolor;
  /* don't move the calculation of pc up here, it depends on which pixels
     are readable */
  for(yo = 0; yo < im->ysize; yo++) {
    for(xo = 0; xo < im->xsize; xo++) {
      for(ch = 0;ch < im->channels; ch++) 
	res[ch] = 0;
      for(c = 0;c < len; c++) {
	i_img_dim xi = xo + c - center;
	if (xi < 0)
	  xi = 0;
	else if (xi >= im->xsize)
	  xi = im->xsize - 1;

	if (i_gpix(im, xi, yo, &rcolor)!=-1) {
	  for(ch = 0; ch < im->channels; ch++) 
            res[ch] += (rcolor.channel[ch])  *coeff[c];
	}
      }
      im_assert(pc != 0);
      for(ch = 0; ch < im->channels; ch++) {
        double temp = res[ch] / pc;
        rcolor.channel[ch] = 
          temp < 0 ? 0 : temp > 255 ? 255 : (i_sample_t)temp;
      }
      i_ppix(timg, xo, yo, &rcolor);
    }
  }

  for(xo = 0; xo < im->xsize; xo++) {
    for(yo = 0;yo < im->ysize; yo++) {
      for(ch =  0; ch < im->channels; ch++)
	res[ch] = 0;
      for(c = 0; c < len; c++) {
	i_img_dim yi = yo + c - center;
	if (yi < 0)
	  yi = 0;
	else if (yi >= im->ysize)
	  yi = im->ysize - 1;
	if (i_gpix(timg, xo, yi, &rcolor) != -1) {
	  for(ch = 0;ch < im->channels; ch++) 
	    res[ch] += (rcolor.channel[ch]) * coeff[c];
	}
      }
      im_assert(pc != 0);
      for(ch = 0;ch < im->channels; ch++) {
	double temp = res[ch] / pc;
	rcolor.channel[ch] = 
	  temp < 0 ? 0 : temp > 255 ? 255 : (i_sample_t)temp;
      }
      i_ppix(im, xo, yo,&rcolor);
    }
  }
  }
  else {
#undef IM_EIGHT_BIT
#undef IM_FILL_COMBINE
#define IM_FILL_COMBINE(fill) ((fill)->combinef)
#undef IM_FILL_FILLER
#define IM_FILL_FILLER(fill) ((fill)->f_fill_with_fcolor)
#line 47 "conv.im"
  i_fcolor rcolor;
  /* don't move the calculation of pc up here, it depends on which pixels
     are readable */
  for(yo = 0; yo < im->ysize; yo++) {
    for(xo = 0; xo < im->xsize; xo++) {
      for(ch = 0;ch < im->channels; ch++) 
	res[ch] = 0;
      for(c = 0;c < len; c++) {
	i_img_dim xi = xo + c - center;
	if (xi < 0)
	  xi = 0;
	else if (xi >= im->xsize)
	  xi = im->xsize - 1;

	if (i_gpixf(im, xi, yo, &rcolor)!=-1) {
	  for(ch = 0; ch < im->channels; ch++) 
            res[ch] += (rcolor.channel[ch])  *coeff[c];
	}
      }
      im_assert(pc != 0);
      for(ch = 0; ch < im->channels; ch++) {
        double temp = res[ch] / pc;
        rcolor.channel[ch] = 
          temp < 0 ? 0 : temp > 1.0 ? 1.0 : (i_fsample_t)temp;
      }
      i_ppixf(timg, xo, yo, &rcolor);
    }
  }

  for(xo = 0; xo < im->xsize; xo++) {
    for(yo = 0;yo < im->ysize; yo++) {
      for(ch =  0; ch < im->channels; ch++)
	res[ch] = 0;
      for(c = 0; c < len; c++) {
	i_img_dim yi = yo + c - center;
	if (yi < 0)
	  yi = 0;
	else if (yi >= im->ysize)
	  yi = im->ysize - 1;
	if (i_gpixf(timg, xo, yi, &rcolor) != -1) {
	  for(ch = 0;ch < im->channels; ch++) 
	    res[ch] += (rcolor.channel[ch]) * coeff[c];
	}
      }
      im_assert(pc != 0);
      for(ch = 0;ch < im->channels; ch++) {
	double temp = res[ch] / pc;
	rcolor.channel[ch] = 
	  temp < 0 ? 0 : temp > 1.0 ? 1.0 : (i_fsample_t)temp;
      }
      i_ppixf(im, xo, yo,&rcolor);
    }
  }
  }
#line 101 "conv.im"

  i_img_destroy(timg);

  return 1;
}
