#define IM_ROUND_8(x) ((int)((x)+0.5))
#define IM_ROUND_double(x) (x)
#define IM_LIMIT_8(x) ((x) < 0 ? 0 : (x) > 255 ? 255 : (x))
#define IM_LIMIT_double(x) ((x) < 0.0 ? 0.0 : (x) > 1.0 ? 1.0 : (x))
#line 1 "flip.im"
#define IMAGER_NO_CONTEXT
#include "imager.h"

static void flip_h(i_img *im);
static void flip_v(i_img *im);
static void flip_hv(i_img *im);

#define XAXIS 0
#define YAXIS 1
#define XYAXIS 2

/*
=item i_flipxy(im, axis)

Flips the image inplace around the axis specified.
Returns 0 if parameters are invalid.

   im   - Image pointer
   axis - 0 = x, 1 = y, 2 = both

=cut
*/

undef_int
i_flipxy(i_img *im, int direction) {
  dIMCTXim(im);
  i_clear_error();

  im_log((aIMCTX, 1, "i_flipxy(im %p, direction %d)\n", im, direction ));

  switch (direction) {
  case XAXIS: /* Horizontal flip */
    flip_h(im);
    break;

  case YAXIS: /* Vertical flip */
    flip_v(im);
    break;

  case XYAXIS: /* Horizontal and Vertical flip */
    flip_hv(im);
    break;

  default:
    im_log((aIMCTX, 1, "i_flipxy: direction is invalid\n" ));
    im_push_errorf(aIMCTX, 0, "direction %d invalid", direction);
    return 0;
  }
  return 1;
}

static void
flip_row_pal(i_palidx *row, i_img_dim width) {
  i_palidx tmp;
  i_palidx *leftp = row;
  i_palidx *rightp = row + width - 1;
  
  while (leftp < rightp) {
    tmp = *leftp;
    *leftp = *rightp;
    *rightp = tmp;
    ++leftp;
    --rightp;
  }
}

#undef IM_EIGHT_BIT
#define IM_EIGHT_BIT 1
#undef IM_FILL_COMBINE
#define IM_FILL_COMBINE(fill) ((fill)->combine)
#undef IM_FILL_FILLER
#define IM_FILL_FILLER(fill) ((fill)->f_fill_with_color)
#line 68 "flip.im"

static void
flip_row_8(i_color *row, i_img_dim width) {
  i_color tmp;
  i_color *leftp = row;
  i_color *rightp = row + width - 1;
  
  while (leftp < rightp) {
    tmp = *leftp;
    *leftp = *rightp;
    *rightp = tmp;
    ++leftp;
    --rightp;
  }
}

#undef IM_EIGHT_BIT
#undef IM_FILL_COMBINE
#define IM_FILL_COMBINE(fill) ((fill)->combinef)
#undef IM_FILL_FILLER
#define IM_FILL_FILLER(fill) ((fill)->f_fill_with_fcolor)
#line 68 "flip.im"

static void
flip_row_double(i_fcolor *row, i_img_dim width) {
  i_fcolor tmp;
  i_fcolor *leftp = row;
  i_fcolor *rightp = row + width - 1;
  
  while (leftp < rightp) {
    tmp = *leftp;
    *leftp = *rightp;
    *rightp = tmp;
    ++leftp;
    --rightp;
  }
}

#line 85 "flip.im"

static void
flip_h(i_img *im) {
  i_img_dim y;
  if (im->type == i_palette_type) {
    i_palidx *line = mymalloc(im->xsize * sizeof(i_palidx));
    for (y = 0; y < im->ysize; ++y) {
      i_gpal(im, 0, im->xsize, y, line);
      flip_row_pal(line, im->xsize);
      i_ppal(im, 0, im->xsize, y, line);
    }
    myfree(line);
  }
  else {
#line 99 "flip.im"
  if (im->bits == i_8_bits) {
#undef IM_EIGHT_BIT
#define IM_EIGHT_BIT 1
#undef IM_FILL_COMBINE
#define IM_FILL_COMBINE(fill) ((fill)->combine)
#undef IM_FILL_FILLER
#define IM_FILL_FILLER(fill) ((fill)->f_fill_with_color)
#line 100 "flip.im"
    i_color *line = mymalloc(im->xsize * sizeof(i_color));
    for (y = 0; y < im->ysize; ++y) {
      i_glin(im, 0, im->xsize, y, line);
      flip_row_8(line, im->xsize);
      i_plin(im, 0, im->xsize, y, line);
    }
    myfree(line);
  }
  else {
#undef IM_EIGHT_BIT
#undef IM_FILL_COMBINE
#define IM_FILL_COMBINE(fill) ((fill)->combinef)
#undef IM_FILL_FILLER
#define IM_FILL_FILLER(fill) ((fill)->f_fill_with_fcolor)
#line 100 "flip.im"
    i_fcolor *line = mymalloc(im->xsize * sizeof(i_fcolor));
    for (y = 0; y < im->ysize; ++y) {
      i_glinf(im, 0, im->xsize, y, line);
      flip_row_double(line, im->xsize);
      i_plinf(im, 0, im->xsize, y, line);
    }
    myfree(line);
  }
#line 108 "flip.im"
  }
}

static void
flip_v(i_img *im) {
  i_img_dim topy = 0;
  i_img_dim boty = im->ysize - 1;
  if (im->type == i_palette_type) {
    i_palidx *top_line = mymalloc(im->xsize * sizeof(i_palidx));
    i_palidx *bot_line = mymalloc(im->xsize * sizeof(i_palidx));
    while (topy < boty) {
      i_gpal(im, 0, im->xsize, topy, top_line);
      i_gpal(im, 0, im->xsize, boty, bot_line);
      i_ppal(im, 0, im->xsize, topy, bot_line);
      i_ppal(im, 0, im->xsize, boty, top_line);
      ++topy;
      --boty;
    }
    myfree(bot_line);
    myfree(top_line);
  }
  else {
#line 130 "flip.im"
  if (im->bits == i_8_bits) {
#undef IM_EIGHT_BIT
#define IM_EIGHT_BIT 1
#undef IM_FILL_COMBINE
#define IM_FILL_COMBINE(fill) ((fill)->combine)
#undef IM_FILL_FILLER
#define IM_FILL_FILLER(fill) ((fill)->f_fill_with_color)
#line 131 "flip.im"
    i_color *top_line = mymalloc(im->xsize * sizeof(i_color));
    i_color *bot_line = mymalloc(im->xsize * sizeof(i_color));
    while (topy < boty) {
      i_glin(im, 0, im->xsize, topy, top_line);
      i_glin(im, 0, im->xsize, boty, bot_line);
      i_plin(im, 0, im->xsize, topy, bot_line);
      i_plin(im, 0, im->xsize, boty, top_line);
      ++topy;
      --boty;
    }
    myfree(top_line);
    myfree(bot_line);
  }
  else {
#undef IM_EIGHT_BIT
#undef IM_FILL_COMBINE
#define IM_FILL_COMBINE(fill) ((fill)->combinef)
#undef IM_FILL_FILLER
#define IM_FILL_FILLER(fill) ((fill)->f_fill_with_fcolor)
#line 131 "flip.im"
    i_fcolor *top_line = mymalloc(im->xsize * sizeof(i_fcolor));
    i_fcolor *bot_line = mymalloc(im->xsize * sizeof(i_fcolor));
    while (topy < boty) {
      i_glinf(im, 0, im->xsize, topy, top_line);
      i_glinf(im, 0, im->xsize, boty, bot_line);
      i_plinf(im, 0, im->xsize, topy, bot_line);
      i_plinf(im, 0, im->xsize, boty, top_line);
      ++topy;
      --boty;
    }
    myfree(top_line);
    myfree(bot_line);
  }
#line 144 "flip.im"
  }
}

static void
flip_hv(i_img *im) {
  i_img_dim topy = 0;
  i_img_dim boty = im->ysize - 1;
  if (im->type == i_palette_type) {
    i_palidx *top_line = mymalloc(im->xsize * sizeof(i_palidx));
    i_palidx *bot_line = mymalloc(im->xsize * sizeof(i_palidx));
    while (topy < boty) {
      i_gpal(im, 0, im->xsize, topy, top_line);
      i_gpal(im, 0, im->xsize, boty, bot_line);
      flip_row_pal(top_line, im->xsize);
      flip_row_pal(bot_line, im->xsize);
      i_ppal(im, 0, im->xsize, topy, bot_line);
      i_ppal(im, 0, im->xsize, boty, top_line);
      ++topy;
      --boty;
    }
    myfree(bot_line);
    myfree(top_line);
  }
  else {
#line 168 "flip.im"
  if (im->bits == i_8_bits) {
#undef IM_EIGHT_BIT
#define IM_EIGHT_BIT 1
#undef IM_FILL_COMBINE
#define IM_FILL_COMBINE(fill) ((fill)->combine)
#undef IM_FILL_FILLER
#define IM_FILL_FILLER(fill) ((fill)->f_fill_with_color)
#line 169 "flip.im"
    i_color *top_line = mymalloc(im->xsize * sizeof(i_color));
    i_color *bot_line = mymalloc(im->xsize * sizeof(i_color));
    while (topy < boty) {
      i_glin(im, 0, im->xsize, topy, top_line);
      i_glin(im, 0, im->xsize, boty, bot_line);
      flip_row_8(top_line, im->xsize);
      flip_row_8(bot_line, im->xsize);
      i_plin(im, 0, im->xsize, topy, bot_line);
      i_plin(im, 0, im->xsize, boty, top_line);
      ++topy;
      --boty;
    }
    if (topy == boty) {
      i_glin(im, 0, im->xsize, topy, top_line);
      flip_row_8(top_line, im->xsize);
      i_plin(im, 0, im->xsize, topy, top_line);
    }
    myfree(top_line);
    myfree(bot_line);
  }
  else {
#undef IM_EIGHT_BIT
#undef IM_FILL_COMBINE
#define IM_FILL_COMBINE(fill) ((fill)->combinef)
#undef IM_FILL_FILLER
#define IM_FILL_FILLER(fill) ((fill)->f_fill_with_fcolor)
#line 169 "flip.im"
    i_fcolor *top_line = mymalloc(im->xsize * sizeof(i_fcolor));
    i_fcolor *bot_line = mymalloc(im->xsize * sizeof(i_fcolor));
    while (topy < boty) {
      i_glinf(im, 0, im->xsize, topy, top_line);
      i_glinf(im, 0, im->xsize, boty, bot_line);
      flip_row_double(top_line, im->xsize);
      flip_row_double(bot_line, im->xsize);
      i_plinf(im, 0, im->xsize, topy, bot_line);
      i_plinf(im, 0, im->xsize, boty, top_line);
      ++topy;
      --boty;
    }
    if (topy == boty) {
      i_glinf(im, 0, im->xsize, topy, top_line);
      flip_row_double(top_line, im->xsize);
      i_plinf(im, 0, im->xsize, topy, top_line);
    }
    myfree(top_line);
    myfree(bot_line);
  }
#line 189 "flip.im"
  }
}
