#define IM_ROUND_8(x) ((int)((x)+0.5))
#define IM_ROUND_double(x) (x)
#define IM_LIMIT_8(x) ((x) < 0 ? 0 : (x) > 255 ? 255 : (x))
#define IM_LIMIT_double(x) ((x) < 0.0 ? 0.0 : (x) > 1.0 ? 1.0 : (x))
#line 1 "render.im"
/*
Render utilities
*/
#include "imager.h"

#define RENDER_MAGIC 0x765AE

typedef void (*render_color_f)(i_render *, i_img_dim, i_img_dim, i_img_dim, unsigned char const *src, i_color const *color);

#define i_has_alpha(channels) ((channels) == 2 || (channels) == 4)

#define i_color_channels(channels) (i_has_alpha(channels) ? (channels)-1 : (channels))

#undef IM_EIGHT_BIT
#define IM_EIGHT_BIT 1
#undef IM_FILL_COMBINE
#define IM_FILL_COMBINE(fill) ((fill)->combine)
#undef IM_FILL_FILLER
#define IM_FILL_FILLER(fill) ((fill)->f_fill_with_color)
#line 15 "render.im"

static void render_color_alpha_8(i_render *r, i_img_dim x, i_img_dim y, i_img_dim width, unsigned char const *src, i_color const *color);
static void render_color_13_8(i_render *r, i_img_dim x, i_img_dim y, i_img_dim width, unsigned char const *src, i_color const *color);

static render_color_f render_color_tab_8[] =
  {
    NULL,
    render_color_13_8,
    render_color_alpha_8,
    render_color_13_8,
    render_color_alpha_8,
  };

static void combine_line_noalpha_8(i_color *out, i_color const *in, int channels, i_img_dim count);
static void combine_line_alpha_8(i_color *out, i_color const *in, int channels, i_img_dim count);
/* the copy variant copies the source alpha to the the output alpha channel */
static void combine_line_alpha_na_8(i_color *out, i_color const *in, int channels, i_img_dim count);

static void combine_line_8(i_color *out, i_color const *in, int channels, i_img_dim count);
static void combine_line_na_8(i_color *out, i_color const *in, int channels, i_img_dim count);

#undef IM_EIGHT_BIT
#undef IM_FILL_COMBINE
#define IM_FILL_COMBINE(fill) ((fill)->combinef)
#undef IM_FILL_FILLER
#define IM_FILL_FILLER(fill) ((fill)->f_fill_with_fcolor)
#line 15 "render.im"

static void render_color_alpha_double(i_render *r, i_img_dim x, i_img_dim y, i_img_dim width, unsigned char const *src, i_color const *color);
static void render_color_13_double(i_render *r, i_img_dim x, i_img_dim y, i_img_dim width, unsigned char const *src, i_color const *color);

static render_color_f render_color_tab_double[] =
  {
    NULL,
    render_color_13_double,
    render_color_alpha_double,
    render_color_13_double,
    render_color_alpha_double,
  };

static void combine_line_noalpha_double(i_fcolor *out, i_fcolor const *in, int channels, i_img_dim count);
static void combine_line_alpha_double(i_fcolor *out, i_fcolor const *in, int channels, i_img_dim count);
/* the copy variant copies the source alpha to the the output alpha channel */
static void combine_line_alpha_na_double(i_fcolor *out, i_fcolor const *in, int channels, i_img_dim count);

static void combine_line_double(i_fcolor *out, i_fcolor const *in, int channels, i_img_dim count);
static void combine_line_na_double(i_fcolor *out, i_fcolor const *in, int channels, i_img_dim count);

#line 37 "render.im"

/* 
=item i_render_new(im, width)
=category Blit tools

Allocate a new C<i_render> object and initialize it.

=cut
*/

i_render *
i_render_new(i_img *im, i_img_dim width) {
  i_render *r = mymalloc(sizeof(i_render));

  i_render_init(r, im, width);

  return r;
}

/*
=item i_render_delete(r)
=category Blit tools

Release an C<i_render> object.

=cut
*/

void
i_render_delete(i_render *r) {
  i_render_done(r);
  myfree(r);
}

void
i_render_init(i_render *r, i_img *im, i_img_dim width) {
  r->magic = RENDER_MAGIC;
  r->im = im;
  r->line_width = width;
  r->line_8 = NULL;
  r->line_double = NULL;
  r->fill_width = width;
  r->fill_line_8 = NULL;
  r->fill_line_double = NULL;
}

void
i_render_done(i_render *r) {
  if (r->line_8)
    myfree(r->line_8);
  if (r->line_double)
    myfree(r->line_double);
  if (r->fill_line_8)
    myfree(r->fill_line_8);
  if (r->fill_line_double)
    myfree(r->fill_line_double);
  r->magic = 0;
}

static void
alloc_line(i_render *r, i_img_dim width, i_img_dim eight_bit) {
  if (width > r->line_width) {
    i_img_dim new_width = r->line_width * 2;
    if (new_width < width)
      new_width = width;

    if (eight_bit) {
      if (r->line_8)
	r->line_8 = myrealloc(r->line_8, sizeof(i_color) * new_width);
      else
	r->line_8 = mymalloc(sizeof(i_color) * new_width);
      if (r->line_double) {
	myfree(r->line_double);
	r->line_double = NULL;
      }
    }
    else {
      if (r->line_double)
	r->line_double = myrealloc(r->line_double, sizeof(i_fcolor) * new_width);
      else
	r->line_double = mymalloc(sizeof(i_fcolor) * new_width);
      if (r->line_8) {
	myfree(r->line_8);
	r->line_double = NULL;
      }
    }

    r->line_width = new_width;
  }
  else {
    if (eight_bit) {
      if (!r->line_8)
	r->line_8 = mymalloc(sizeof(i_color) * r->line_width);
      if (r->line_double) {
	myfree(r->line_double);
	r->line_double = NULL;
      }
    }
    else {
      if (!r->line_double)
	r->line_double = mymalloc(sizeof(i_fcolor) * r->line_width);
      if (r->line_8) {
	myfree(r->line_8);
	r->line_8 = NULL;
      }
    }
  }
}

static void
alloc_fill_line(i_render *r, i_img_dim width, int eight_bit) {
  if (width > r->fill_width) {
    i_img_dim new_width = r->fill_width * 2;
    if (new_width < width)
      new_width = width;

    if (eight_bit) {
      if (r->line_8)
	r->fill_line_8 = myrealloc(r->fill_line_8, sizeof(i_color) * new_width);
      else
	r->fill_line_8 = mymalloc(sizeof(i_color) * new_width);
      if (r->fill_line_double) {
	myfree(r->fill_line_double);
	r->fill_line_double = NULL;
      }
    }
    else {
      if (r->fill_line_double)
	r->fill_line_double = myrealloc(r->fill_line_double, sizeof(i_fcolor) * new_width);
      else
	r->fill_line_double = mymalloc(sizeof(i_fcolor) * new_width);
      if (r->fill_line_8) {
	myfree(r->fill_line_8);
	r->fill_line_double = NULL;
      }
    }

    r->fill_width = new_width;
  }
  else {
    if (eight_bit) {
      if (!r->fill_line_8)
	r->fill_line_8 = mymalloc(sizeof(i_color) * r->fill_width);
      if (r->fill_line_double) {
	myfree(r->fill_line_double);
	r->fill_line_double = NULL;
      }
    }
    else {
      if (!r->fill_line_double)
	r->fill_line_double = mymalloc(sizeof(i_fcolor) * r->fill_width);
      if (r->fill_line_8) {
	myfree(r->fill_line_8);
	r->fill_line_8 = NULL;
      }
    }
  }
}

/*
=item i_render_color(r, x, y, width, source, color)
=category Blit tools

Render the given color with the coverage specified by C<source[0]> to
C<source[width-1]>.

Renders in normal combine mode.

=cut
*/

void
i_render_color(i_render *r, i_img_dim x, i_img_dim y, i_img_dim width,
	       unsigned char const *src, i_color const *color) {
  i_img *im = r->im;
  if (y < 0 || y >= im->ysize)
    return;
  if (x < 0) {
    width += x;
    src -= x;
    x = 0;
  }
  if (x + width > im->xsize) {
    width = im->xsize - x;
  }
  if (x >= im->xsize || x + width <= 0 || width <= 0)
    return;

  /* avoid as much work as we can */
  while (width > 0 && *src == 0) {
    --width;
    ++src;
    ++x;
  }
  while (width > 0 && src[width-1] == 0) {
    --width;
  }
  if (!width)
    return;

  alloc_line(r, width, r->im->bits <= 8);

#line 239 "render.im"
  if (r->im->bits <= 8) {
#undef IM_EIGHT_BIT
#define IM_EIGHT_BIT 1
#undef IM_FILL_COMBINE
#define IM_FILL_COMBINE(fill) ((fill)->combine)
#undef IM_FILL_FILLER
#define IM_FILL_FILLER(fill) ((fill)->f_fill_with_color)
#line 240 "render.im"
  /*if (r->line_8 == NULL)
    r->line_8 = mymalloc(sizeof(i_color) * r->width);*/
  (render_color_tab_8[im->channels])(r, x, y, width, src, color);
  }
  else {
#undef IM_EIGHT_BIT
#undef IM_FILL_COMBINE
#define IM_FILL_COMBINE(fill) ((fill)->combinef)
#undef IM_FILL_FILLER
#define IM_FILL_FILLER(fill) ((fill)->f_fill_with_fcolor)
#line 240 "render.im"
  /*if (r->line_double == NULL)
    r->line_double = mymalloc(sizeof(i_fcolor) * r->width);*/
  (render_color_tab_double[im->channels])(r, x, y, width, src, color);
  }
#line 244 "render.im"
}

/*
=item i_render_fill(r, x, y, width, source, fill)
=category Blit tools

Render the given fill with the coverage in C<source[0]> through
C<source[width-1]>.

=cut
*/

void
i_render_fill(i_render *r, i_img_dim x, i_img_dim y, i_img_dim width,
	      unsigned char const *src, i_fill_t *fill) {
  i_img *im = r->im;
  int fill_channels = im->channels;
  
  if (fill_channels == 1 || fill_channels == 3)
    ++fill_channels;

  if (y < 0 || y >= im->ysize)
    return;
  if (x < 0) {
    width += x;
    src -= x;
    x = 0;
  }
  if (x + width > im->xsize) {
    width = im->xsize - x;
  }
  if (x >= im->xsize || x + width <= 0 || width <= 0)
    return;

  if (src) {
    /* avoid as much work as we can */
    while (width > 0 && *src == 0) {
      --width;
      ++src;
      ++x;
    }
    while (width > 0 && src[width-1] == 0) {
      --width;
    }
  }
  if (!width)
    return;

  alloc_line(r, width, r->im->bits <= 8 && fill->f_fill_with_color != NULL);
  alloc_fill_line(r, width, r->im->bits <= 8 && fill->f_fill_with_color != NULL);

#line 295 "render.im"
  if (r->im->bits <= 8 && fill->f_fill_with_color) {
#undef IM_EIGHT_BIT
#define IM_EIGHT_BIT 1
#undef IM_FILL_COMBINE
#define IM_FILL_COMBINE(fill) ((fill)->combine)
#undef IM_FILL_FILLER
#define IM_FILL_FILLER(fill) ((fill)->f_fill_with_color)
#line 296 "render.im"
  if (IM_FILL_COMBINE(fill)) {
    i_color *srcc = r->fill_line_8;
    i_color *destc = r->line_8;
    IM_FILL_FILLER(fill)(fill, x, y, width, fill_channels, r->fill_line_8);
    if (src) {
      unsigned char const *srcc = src;
      i_color *fillc = r->fill_line_8;
      i_img_dim work_width = width;
      while (work_width) {
	if (*srcc == 0) {
	  fillc->channel[fill_channels-1] = 0;
	}
	else if (*srcc != 255) {
	  fillc->channel[fill_channels-1] =
	    fillc->channel[fill_channels-1] * *srcc / 255;
	}
	--work_width;
	++srcc;
	++fillc;
      }
    }
    i_glin(r->im, x, x+width, y, r->line_8);
    IM_FILL_COMBINE(fill)(destc, srcc, r->im->channels, width);
  }
  else {
    if (src) {
      i_img_dim work_width = width;
      i_color *srcc = r->fill_line_8;
      i_color *destc = r->line_8;
      int ch;

      IM_FILL_FILLER(fill)(fill, x, y, width, fill_channels, r->fill_line_8);
      i_glin(r->im, x, x+width, y, r->line_8);
      while (work_width) {
	if (*src == 255) {
	  /* just replace it */
	  *destc = *srcc;
	}
	else if (*src) {
	  for (ch = 0; ch < im->channels; ++ch) {
	    int work = (destc->channel[ch] * (255 - *src)
			      + srcc->channel[ch] * *src) / 255.0;
	    destc->channel[ch] = IM_LIMIT_8(work);
	  }
	}
	
	++srcc;
	++destc;
	++src;
	--work_width;
      }
    }
    else { /* if (src) */
      IM_FILL_FILLER(fill)(fill, x, y, width, fill_channels, r->line_8);
    }
  }
  i_plin(im, x, x+width, y, r->line_8);
  }
  else {
#undef IM_EIGHT_BIT
#undef IM_FILL_COMBINE
#define IM_FILL_COMBINE(fill) ((fill)->combinef)
#undef IM_FILL_FILLER
#define IM_FILL_FILLER(fill) ((fill)->f_fill_with_fcolor)
#line 296 "render.im"
  if (IM_FILL_COMBINE(fill)) {
    i_fcolor *srcc = r->fill_line_double;
    i_fcolor *destc = r->line_double;
    IM_FILL_FILLER(fill)(fill, x, y, width, fill_channels, r->fill_line_double);
    if (src) {
      unsigned char const *srcc = src;
      i_fcolor *fillc = r->fill_line_double;
      i_img_dim work_width = width;
      while (work_width) {
	if (*srcc == 0) {
	  fillc->channel[fill_channels-1] = 0;
	}
	else if (*srcc != 255) {
	  fillc->channel[fill_channels-1] =
	    fillc->channel[fill_channels-1] * *srcc / 255;
	}
	--work_width;
	++srcc;
	++fillc;
      }
    }
    i_glinf(r->im, x, x+width, y, r->line_double);
    IM_FILL_COMBINE(fill)(destc, srcc, r->im->channels, width);
  }
  else {
    if (src) {
      i_img_dim work_width = width;
      i_fcolor *srcc = r->fill_line_double;
      i_fcolor *destc = r->line_double;
      int ch;

      IM_FILL_FILLER(fill)(fill, x, y, width, fill_channels, r->fill_line_double);
      i_glinf(r->im, x, x+width, y, r->line_double);
      while (work_width) {
	if (*src == 255) {
	  /* just replace it */
	  *destc = *srcc;
	}
	else if (*src) {
	  for (ch = 0; ch < im->channels; ++ch) {
	    double work = (destc->channel[ch] * (255 - *src)
			      + srcc->channel[ch] * *src) / 255.0;
	    destc->channel[ch] = IM_LIMIT_double(work);
	  }
	}
	
	++srcc;
	++destc;
	++src;
	--work_width;
      }
    }
    else { /* if (src) */
      IM_FILL_FILLER(fill)(fill, x, y, width, fill_channels, r->line_double);
    }
  }
  i_plinf(im, x, x+width, y, r->line_double);
  }
#line 354 "render.im"
}

#if 0

/* for debuggin */

static void
dump_src(const char *note, unsigned char const *src, i_img_dim width) {
  i_img_dim i;
  printf("%s - %p/%" i_DF "\n", note, src, i_DFc(width));
  for (i = 0; i < width; ++i) {
    printf("%02x ", src[i]);
  }
  putchar('\n');
}

#endif

#undef IM_EIGHT_BIT
#define IM_EIGHT_BIT 1
#undef IM_FILL_COMBINE
#define IM_FILL_COMBINE(fill) ((fill)->combine)
#undef IM_FILL_FILLER
#define IM_FILL_FILLER(fill) ((fill)->f_fill_with_color)
#line 373 "render.im"

/*
=item i_render_line(r, x, y, width, source, fill)
=category Blit tools

Render the given fill with the coverage in C<source[0]> through
C<source[width-1]>.

=cut

=item i_render_linef(r, x, y, width, source, fill)
=category Blit tools

Render the given fill with the coverage in C<source[0]> through
C<source[width-1]>.

=cut
*/

void
i_render_line(i_render *r, i_img_dim x, i_img_dim y, i_img_dim width,
	       const i_sample_t *src, i_color *line,
	       i_fill_combine_f combine) {
  i_img *im = r->im;
  int src_chans = im->channels;

  /* src must always have an alpha channel */
  if (src_chans == 1 || src_chans == 3)
    ++src_chans;

  if (y < 0 || y >= im->ysize)
    return;
  if (x < 0) {
    src -= x;
    line -= x;
    width += x;
    x = 0;
  }
  if (x + width > im->xsize)
    width = r->im->xsize - x;

#ifdef IM_EIGHT_BIT
  alloc_line(r, width, 1);
#else
  alloc_line(r, width, 0);
#endif

  if (combine) {
    if (src) {
      i_img_dim work_width = width;
      i_color *linep = line;
      const i_sample_t *srcp = src;
      int alpha_chan = src_chans - 1;
      
      while (work_width) {
	if (*srcp) {
	  if (*srcp != 255) 
	    linep->channel[alpha_chan] = 
	      linep->channel[alpha_chan] * *srcp / 255;
	}
	else {
	  linep->channel[alpha_chan] = 0;
	}
	--work_width;
	++srcp;
	++linep;
      }
    }
    i_glin(im, x, x+width, y, r->line_8);
    combine(r->line_8, line, im->channels, width);
    i_plin(im, x, x+width, y, r->line_8);
  }
  else {
    if (src) {
      i_img_dim work_width = width;
      i_color *srcc = line;
      i_color *destc = r->line_8;

      i_glin(im, x, x+width, y, r->line_8);
      while (work_width) {
	if (*src == 255) {
	  /* just replace it */
	  *destc = *srcc;
	}
	else if (*src) {
	  int ch;
	  for (ch = 0; ch < im->channels; ++ch) {
	    int work = (destc->channel[ch] * (255 - *src)
			      + srcc->channel[ch] * *src) / 255;
	    destc->channel[ch] = IM_LIMIT_8(work);
	  }
	}
	
	++srcc;
	++destc;
	++src;
	--work_width;
      }
      i_plin(im, x, x+width, y, r->line_8);
    }
    else {
      i_plin(im, x, x+width, y, line);
    }
  }
}

static
void
render_color_13_8(i_render *r, i_img_dim x, i_img_dim y,
			   i_img_dim width, unsigned char const *src,
			   i_color const *color) {
  i_img *im = r->im;
  i_color *linep = r->line_8;
  int ch, channels = im->channels;
  i_img_dim fetch_offset;
  int color_alpha = color->channel[im->channels];
#undef STORE_COLOR
#ifdef IM_EIGHT_BIT
#define STORE_COLOR (*color)
#else
  i_fcolor fcolor;

  for (ch = 0; ch < channels; ++ch) {
    fcolor.channel[ch] = color->channel[ch] / 255.0;
  }
#define STORE_COLOR fcolor
#endif
 
  fetch_offset = 0;
  if (color_alpha == 0xFF) {
    while (fetch_offset < width && *src == 0xFF) {
      *linep++ = STORE_COLOR;
      ++src;
      ++fetch_offset;
    }
  }
  i_glin(im, x+fetch_offset, x+width, y, linep);
  while (fetch_offset < width) {
#ifdef IM_EIGHT_BIT
    int alpha = *src++ * color_alpha / 255;
#else
    int alpha = *src++ * color_alpha / (255.0 * 255.0);
#endif
    if (alpha == 255)
      *linep = STORE_COLOR;
    else if (alpha) {
      for (ch = 0; ch < channels; ++ch) {
        linep->channel[ch] = (linep->channel[ch] * (255 - alpha) 
                              + STORE_COLOR.channel[ch] * alpha) / 255;
      }
    }
    ++linep;
    ++fetch_offset;
  }
  i_plin(im, x, x+width, y, r->line_8);
}

static
void
render_color_alpha_8(i_render *r, i_img_dim x, i_img_dim y,
			      i_img_dim width, unsigned char const *src,
			      i_color const *color) {
  i_color *linep = r->line_8;
  int ch;
  int alpha_channel = r->im->channels - 1;
  i_img_dim fetch_offset;
  int color_alpha = color->channel[alpha_channel];
#undef STORE_COLOR
#ifdef IM_EIGHT_BIT
#define STORE_COLOR (*color)
#else
  i_fcolor fcolor;

  for (ch = 0; ch < r->im->channels; ++ch) {
    fcolor.channel[ch] = color->channel[ch] / 255.0;
  }
#define STORE_COLOR fcolor
#endif

  fetch_offset = 0;
  if (color->channel[alpha_channel] == 0xFF) {
    while (fetch_offset < width && *src == 0xFF) {
      *linep++ = STORE_COLOR;
      ++src;
      ++fetch_offset;
    }
  }
  i_glin(r->im, x+fetch_offset, x+width, y, linep);
  while (fetch_offset < width) {
#ifdef IM_EIGHT_BIT
    int src_alpha = *src++ * color_alpha / 255;
#else
    int src_alpha = *src++ * color_alpha / (255.0 * 255.0);
#endif
    if (src_alpha == 255)
      *linep = STORE_COLOR;
    else if (src_alpha) {
      int remains = 255 - src_alpha;
      int orig_alpha = linep->channel[alpha_channel];
      int dest_alpha = src_alpha + (remains * orig_alpha) / 255;
      for (ch = 0; ch < alpha_channel; ++ch) {
        linep->channel[ch] = ( src_alpha * STORE_COLOR.channel[ch]
                               + remains * linep->channel[ch] * orig_alpha / 255
                               ) / dest_alpha;
      }
      linep->channel[alpha_channel] = dest_alpha;
    }
    ++linep;
    ++fetch_offset;
  }
  i_plin(r->im, x, x+width, y, r->line_8);
#undef STORE_COLOR
}

/* combine a line of image data with an output line, both the input
   and output lines include an alpha channel.

   Both input and output lines have I<channels> of data, channels
   should be either 2 or 4.
*/

static void
combine_line_alpha_8(i_color *out, i_color const *in, 
			      int channels, i_img_dim count) {
  int ch;
  int alpha_channel = channels - 1;
  
  while (count) {
    int src_alpha = in->channel[alpha_channel];
      
    if (src_alpha == 255)
      *out = *in;
    else if (src_alpha) {
      int remains = 255 - src_alpha;
      int orig_alpha = out->channel[alpha_channel];
      int dest_alpha = src_alpha + (remains * orig_alpha) / 255;
	
      for (ch = 0; ch < alpha_channel; ++ch) {
	out->channel[ch] = ( src_alpha * in->channel[ch]
			     + remains * out->channel[ch] * orig_alpha / 255
			     ) / dest_alpha;
      }
      out->channel[alpha_channel] = dest_alpha;
    }

    ++out;
    ++in;
    --count;
  }
}

/* combine a line of image data with an output line.  The input line
   includes an alpha channel, the output line has no alpha channel.
   
   The input line has I<channels>+1 of color data.  The output line
   has I<channels> of color data.
*/

static void
combine_line_noalpha_8
     (i_color *out, i_color const *in, int channels, i_img_dim count) {
  int ch;

  while (count) {
    int src_alpha = in->channel[channels];
    
    if (src_alpha == 255)
      *out = *in;
    else if (src_alpha) {
      int remains;
      
      remains = 255 - src_alpha;
      for (ch = 0; ch < channels; ++ch) {
	out->channel[ch] = ( in->channel[ch] * src_alpha
			     + out->channel[ch] * remains) / 255;
      }
    }
    
    ++out;
    ++in;
    --count;
  }
}

/* combine a line of image data with an output line, both the input
   and output lines include an alpha channel.

   Both input and output lines have I<channels> of data, channels
   should be either 2 or 4.

   This variant does not modify the output alpha channel.
*/

static void
combine_line_alpha_na_8(i_color *out, i_color const *in, 
				   int channels, i_img_dim count) {
  int ch;
  int alpha_channel = channels - 1;
  
  while (count) {
    int src_alpha = in->channel[alpha_channel];
      
    if (src_alpha == 255)
      *out = *in;
    else if (src_alpha) {
      int remains = 255 - src_alpha;
      int orig_alpha = out->channel[alpha_channel];
      int dest_alpha = src_alpha + (remains * orig_alpha) / 255;
	
      for (ch = 0; ch < alpha_channel; ++ch) {
	out->channel[ch] = ( src_alpha * in->channel[ch]
			     + remains * out->channel[ch] * orig_alpha / 255
			     ) / dest_alpha;
      }
    }

    ++out;
    ++in;
    --count;
  }
}

static void
combine_line_8(i_color *out, i_color const *in, int channels, i_img_dim count) {
  if (channels == 2 || channels == 4)
    combine_line_alpha_8(out, in, channels, count);
  else
    combine_line_noalpha_8(out, in, channels, count);
}

static void
combine_line_na_8(i_color *out, i_color const *in, int channels, i_img_dim count) {
  if (channels == 2 || channels == 4)
    combine_line_alpha_na_8(out, in, channels, count);
  else
    combine_line_noalpha_8(out, in, channels, count);
}

static void combine_alphablend_8(i_color *, i_color *, int, i_img_dim);
static void combine_mult_8(i_color *, i_color *, int, i_img_dim);
static void combine_dissolve_8(i_color *, i_color *, int, i_img_dim);
static void combine_add_8(i_color *, i_color *, int, i_img_dim);
static void combine_subtract_8(i_color *, i_color *, int, i_img_dim);
static void combine_diff_8(i_color *, i_color *, int, i_img_dim);
static void combine_darken_8(i_color *, i_color *, int, i_img_dim);
static void combine_lighten_8(i_color *, i_color *, int, i_img_dim);
static void combine_hue_8(i_color *, i_color *, int, i_img_dim);
static void combine_sat_8(i_color *, i_color *, int, i_img_dim);
static void combine_value_8(i_color *, i_color *, int, i_img_dim);
static void combine_color_8(i_color *, i_color *, int, i_img_dim);

static const i_fill_combine_f combines_8[] =
{
  NULL,
  combine_alphablend_8,
  combine_mult_8,
  combine_dissolve_8,
  combine_add_8,
  combine_subtract_8,
  combine_diff_8,
  combine_lighten_8,
  combine_darken_8,
  combine_hue_8,
  combine_sat_8,
  combine_value_8,
  combine_color_8
};

#undef IM_EIGHT_BIT
#undef IM_FILL_COMBINE
#define IM_FILL_COMBINE(fill) ((fill)->combinef)
#undef IM_FILL_FILLER
#define IM_FILL_FILLER(fill) ((fill)->f_fill_with_fcolor)
#line 373 "render.im"

/*
=item i_render_line(r, x, y, width, source, fill)
=category Blit tools

Render the given fill with the coverage in C<source[0]> through
C<source[width-1]>.

=cut

=item i_render_linef(r, x, y, width, source, fill)
=category Blit tools

Render the given fill with the coverage in C<source[0]> through
C<source[width-1]>.

=cut
*/

void
i_render_linef(i_render *r, i_img_dim x, i_img_dim y, i_img_dim width,
	       const i_fsample_t *src, i_fcolor *line,
	       i_fill_combinef_f combine) {
  i_img *im = r->im;
  int src_chans = im->channels;

  /* src must always have an alpha channel */
  if (src_chans == 1 || src_chans == 3)
    ++src_chans;

  if (y < 0 || y >= im->ysize)
    return;
  if (x < 0) {
    src -= x;
    line -= x;
    width += x;
    x = 0;
  }
  if (x + width > im->xsize)
    width = r->im->xsize - x;

#ifdef IM_EIGHT_BIT
  alloc_line(r, width, 1);
#else
  alloc_line(r, width, 0);
#endif

  if (combine) {
    if (src) {
      i_img_dim work_width = width;
      i_fcolor *linep = line;
      const i_fsample_t *srcp = src;
      int alpha_chan = src_chans - 1;
      
      while (work_width) {
	if (*srcp) {
	  if (*srcp != 1.0) 
	    linep->channel[alpha_chan] = 
	      linep->channel[alpha_chan] * *srcp / 1.0;
	}
	else {
	  linep->channel[alpha_chan] = 0;
	}
	--work_width;
	++srcp;
	++linep;
      }
    }
    i_glinf(im, x, x+width, y, r->line_double);
    combine(r->line_double, line, im->channels, width);
    i_plinf(im, x, x+width, y, r->line_double);
  }
  else {
    if (src) {
      i_img_dim work_width = width;
      i_fcolor *srcc = line;
      i_fcolor *destc = r->line_double;

      i_glinf(im, x, x+width, y, r->line_double);
      while (work_width) {
	if (*src == 255) {
	  /* just replace it */
	  *destc = *srcc;
	}
	else if (*src) {
	  int ch;
	  for (ch = 0; ch < im->channels; ++ch) {
	    double work = (destc->channel[ch] * (1.0 - *src)
			      + srcc->channel[ch] * *src) / 1.0;
	    destc->channel[ch] = IM_LIMIT_double(work);
	  }
	}
	
	++srcc;
	++destc;
	++src;
	--work_width;
      }
      i_plinf(im, x, x+width, y, r->line_double);
    }
    else {
      i_plinf(im, x, x+width, y, line);
    }
  }
}

static
void
render_color_13_double(i_render *r, i_img_dim x, i_img_dim y,
			   i_img_dim width, unsigned char const *src,
			   i_color const *color) {
  i_img *im = r->im;
  i_fcolor *linep = r->line_double;
  int ch, channels = im->channels;
  i_img_dim fetch_offset;
  int color_alpha = color->channel[im->channels];
#undef STORE_COLOR
#ifdef IM_EIGHT_BIT
#define STORE_COLOR (*color)
#else
  i_fcolor fcolor;

  for (ch = 0; ch < channels; ++ch) {
    fcolor.channel[ch] = color->channel[ch] / 255.0;
  }
#define STORE_COLOR fcolor
#endif
 
  fetch_offset = 0;
  if (color_alpha == 0xFF) {
    while (fetch_offset < width && *src == 0xFF) {
      *linep++ = STORE_COLOR;
      ++src;
      ++fetch_offset;
    }
  }
  i_glinf(im, x+fetch_offset, x+width, y, linep);
  while (fetch_offset < width) {
#ifdef IM_EIGHT_BIT
    double alpha = *src++ * color_alpha / 255;
#else
    double alpha = *src++ * color_alpha / (255.0 * 255.0);
#endif
    if (alpha == 1.0)
      *linep = STORE_COLOR;
    else if (alpha) {
      for (ch = 0; ch < channels; ++ch) {
        linep->channel[ch] = (linep->channel[ch] * (1.0 - alpha) 
                              + STORE_COLOR.channel[ch] * alpha) / 1.0;
      }
    }
    ++linep;
    ++fetch_offset;
  }
  i_plinf(im, x, x+width, y, r->line_double);
}

static
void
render_color_alpha_double(i_render *r, i_img_dim x, i_img_dim y,
			      i_img_dim width, unsigned char const *src,
			      i_color const *color) {
  i_fcolor *linep = r->line_double;
  int ch;
  int alpha_channel = r->im->channels - 1;
  i_img_dim fetch_offset;
  int color_alpha = color->channel[alpha_channel];
#undef STORE_COLOR
#ifdef IM_EIGHT_BIT
#define STORE_COLOR (*color)
#else
  i_fcolor fcolor;

  for (ch = 0; ch < r->im->channels; ++ch) {
    fcolor.channel[ch] = color->channel[ch] / 255.0;
  }
#define STORE_COLOR fcolor
#endif

  fetch_offset = 0;
  if (color->channel[alpha_channel] == 0xFF) {
    while (fetch_offset < width && *src == 0xFF) {
      *linep++ = STORE_COLOR;
      ++src;
      ++fetch_offset;
    }
  }
  i_glinf(r->im, x+fetch_offset, x+width, y, linep);
  while (fetch_offset < width) {
#ifdef IM_EIGHT_BIT
    double src_alpha = *src++ * color_alpha / 255;
#else
    double src_alpha = *src++ * color_alpha / (255.0 * 255.0);
#endif
    if (src_alpha == 1.0)
      *linep = STORE_COLOR;
    else if (src_alpha) {
      double remains = 1.0 - src_alpha;
      double orig_alpha = linep->channel[alpha_channel];
      double dest_alpha = src_alpha + (remains * orig_alpha) / 1.0;
      for (ch = 0; ch < alpha_channel; ++ch) {
        linep->channel[ch] = ( src_alpha * STORE_COLOR.channel[ch]
                               + remains * linep->channel[ch] * orig_alpha / 1.0
                               ) / dest_alpha;
      }
      linep->channel[alpha_channel] = dest_alpha;
    }
    ++linep;
    ++fetch_offset;
  }
  i_plinf(r->im, x, x+width, y, r->line_double);
#undef STORE_COLOR
}

/* combine a line of image data with an output line, both the input
   and output lines include an alpha channel.

   Both input and output lines have I<channels> of data, channels
   should be either 2 or 4.
*/

static void
combine_line_alpha_double(i_fcolor *out, i_fcolor const *in, 
			      int channels, i_img_dim count) {
  int ch;
  int alpha_channel = channels - 1;
  
  while (count) {
    double src_alpha = in->channel[alpha_channel];
      
    if (src_alpha == 1.0)
      *out = *in;
    else if (src_alpha) {
      double remains = 1.0 - src_alpha;
      double orig_alpha = out->channel[alpha_channel];
      double dest_alpha = src_alpha + (remains * orig_alpha) / 1.0;
	
      for (ch = 0; ch < alpha_channel; ++ch) {
	out->channel[ch] = ( src_alpha * in->channel[ch]
			     + remains * out->channel[ch] * orig_alpha / 1.0
			     ) / dest_alpha;
      }
      out->channel[alpha_channel] = dest_alpha;
    }

    ++out;
    ++in;
    --count;
  }
}

/* combine a line of image data with an output line.  The input line
   includes an alpha channel, the output line has no alpha channel.
   
   The input line has I<channels>+1 of color data.  The output line
   has I<channels> of color data.
*/

static void
combine_line_noalpha_double
     (i_fcolor *out, i_fcolor const *in, int channels, i_img_dim count) {
  int ch;

  while (count) {
    double src_alpha = in->channel[channels];
    
    if (src_alpha == 1.0)
      *out = *in;
    else if (src_alpha) {
      double remains;
      
      remains = 1.0 - src_alpha;
      for (ch = 0; ch < channels; ++ch) {
	out->channel[ch] = ( in->channel[ch] * src_alpha
			     + out->channel[ch] * remains) / 1.0;
      }
    }
    
    ++out;
    ++in;
    --count;
  }
}

/* combine a line of image data with an output line, both the input
   and output lines include an alpha channel.

   Both input and output lines have I<channels> of data, channels
   should be either 2 or 4.

   This variant does not modify the output alpha channel.
*/

static void
combine_line_alpha_na_double(i_fcolor *out, i_fcolor const *in, 
				   int channels, i_img_dim count) {
  int ch;
  int alpha_channel = channels - 1;
  
  while (count) {
    double src_alpha = in->channel[alpha_channel];
      
    if (src_alpha == 1.0)
      *out = *in;
    else if (src_alpha) {
      double remains = 1.0 - src_alpha;
      double orig_alpha = out->channel[alpha_channel];
      double dest_alpha = src_alpha + (remains * orig_alpha) / 1.0;
	
      for (ch = 0; ch < alpha_channel; ++ch) {
	out->channel[ch] = ( src_alpha * in->channel[ch]
			     + remains * out->channel[ch] * orig_alpha / 1.0
			     ) / dest_alpha;
      }
    }

    ++out;
    ++in;
    --count;
  }
}

static void
combine_line_double(i_fcolor *out, i_fcolor const *in, int channels, i_img_dim count) {
  if (channels == 2 || channels == 4)
    combine_line_alpha_double(out, in, channels, count);
  else
    combine_line_noalpha_double(out, in, channels, count);
}

static void
combine_line_na_double(i_fcolor *out, i_fcolor const *in, int channels, i_img_dim count) {
  if (channels == 2 || channels == 4)
    combine_line_alpha_na_double(out, in, channels, count);
  else
    combine_line_noalpha_double(out, in, channels, count);
}

static void combine_alphablend_double(i_fcolor *, i_fcolor *, int, i_img_dim);
static void combine_mult_double(i_fcolor *, i_fcolor *, int, i_img_dim);
static void combine_dissolve_double(i_fcolor *, i_fcolor *, int, i_img_dim);
static void combine_add_double(i_fcolor *, i_fcolor *, int, i_img_dim);
static void combine_subtract_double(i_fcolor *, i_fcolor *, int, i_img_dim);
static void combine_diff_double(i_fcolor *, i_fcolor *, int, i_img_dim);
static void combine_darken_double(i_fcolor *, i_fcolor *, int, i_img_dim);
static void combine_lighten_double(i_fcolor *, i_fcolor *, int, i_img_dim);
static void combine_hue_double(i_fcolor *, i_fcolor *, int, i_img_dim);
static void combine_sat_double(i_fcolor *, i_fcolor *, int, i_img_dim);
static void combine_value_double(i_fcolor *, i_fcolor *, int, i_img_dim);
static void combine_color_double(i_fcolor *, i_fcolor *, int, i_img_dim);

static const i_fill_combinef_f combines_double[] =
{
  NULL,
  combine_alphablend_double,
  combine_mult_double,
  combine_dissolve_double,
  combine_add_double,
  combine_subtract_double,
  combine_diff_double,
  combine_lighten_double,
  combine_darken_double,
  combine_hue_double,
  combine_sat_double,
  combine_value_double,
  combine_color_double
};

#line 742 "render.im"

/*
=item i_get_combine(combine, color_func, fcolor_func)

=cut
*/

void i_get_combine(int combine, i_fill_combine_f *color_func, 
                   i_fill_combinef_f *fcolor_func) {
  if (combine < 0 || combine >= sizeof(combines_8) / sizeof(*combines_8))
    combine = 0;

  *color_func = combines_8[combine];
  *fcolor_func = combines_double[combine];
}

#undef IM_EIGHT_BIT
#define IM_EIGHT_BIT 1
#undef IM_FILL_COMBINE
#define IM_FILL_COMBINE(fill) ((fill)->combine)
#undef IM_FILL_FILLER
#define IM_FILL_FILLER(fill) ((fill)->f_fill_with_color)
#line 759 "render.im"

/*
  Three good references for implementing combining modes:

  http://www.w3.org/TR/2004/WD-SVG12-20041027/rendering.html
  referenced as [svg1.2]

  http://gimp-savvy.com/BOOK/index.html?node55.html
  ("The Blending Modes", if it changes)
  referenced as [savvy]

  http://www.pegtop.net/delphi/articles/blendmodes/
  referenced as [pegtop]

  Where differences exist, I follow the SVG practice, the gimp
  practice, and lastly pegtop.
*/


static void 
combine_alphablend_8(i_color *out, i_color *in, int channels, i_img_dim count) {
  combine_line_8(out, in, channels, count);
}

/*
Dca' = Sca.Dca + Sca.(1 - Da) + Dca.(1 - Sa)
Da'  = Sa.Da + Sa.(1 - Da) + Da.(1 - Sa)
     = Sa + Da - Sa.Da

When Da=1

Dc' = Sc.Sa.Dc + Dc.(1 - Sa)
 */
static void
combine_mult_8(i_color *out, i_color *in, int channels, i_img_dim count) {
  int ch;
  i_color *inp = in;
  i_color *outp = out;
  i_img_dim work_count = count;
  int color_channels = i_color_channels(channels);

  if (i_has_alpha(channels)) {
    while (work_count--) {
      int orig_alpha = outp->channel[color_channels];
      int src_alpha = inp->channel[color_channels];
      
      if (src_alpha) {
	int dest_alpha = src_alpha + orig_alpha 
	  - (src_alpha * orig_alpha) / 255;
	
	for (ch = 0; ch < color_channels; ++ch) { 
	  outp->channel[ch] = 
	    (inp->channel[ch] * src_alpha * outp->channel[ch] / 255
	     * orig_alpha
	     + inp->channel[ch] * src_alpha * (255 - orig_alpha)
	     + outp->channel[ch] * orig_alpha * (255 - src_alpha))
	    / 255 / dest_alpha;
	}
	outp->channel[color_channels] = dest_alpha;
      }
      ++outp;
      ++inp;
    }
  }
  else {
    while (work_count--) {
      int src_alpha = inp->channel[color_channels];
      int remains = 255 - src_alpha;
      
      if (src_alpha) {
	for (ch = 0; ch < color_channels; ++ch) { 
	  outp->channel[ch] = 
	    (src_alpha * inp->channel[ch] * outp->channel[ch] / 255
	     + outp->channel[ch] * remains) / 255;
	}
      }
      ++outp;
      ++inp;
    }
  }
}

static void 
combine_dissolve_8(i_color *out, i_color *in, int channels, i_img_dim count) {
  int color_channels = i_color_channels(channels);
  int ch;

  if (i_has_alpha(channels)) {
    while (count--) {
      if (in->channel[channels-1] > rand() * ((double)255 / RAND_MAX)) {
	for (ch = 0; ch < color_channels; ++ch) {
	  out->channel[ch] = in->channel[ch];
	}
	out->channel[color_channels] = 255;
      }
      ++out;
      ++in;
    }
  }
  else {
    while (count--) {
      if (in->channel[channels] > rand() * ((double)255 / RAND_MAX)) {
	for (ch = 0; ch < color_channels; ++ch) {
	  out->channel[ch] = in->channel[ch];
	}
      }
      ++out;
      ++in;
    }
  }
}

/*
Dca' = Sca.Da + Dca.Sa + Sca.(1 - Da) + Dca.(1 - Sa)
     = Sca + Dca
Da'  = Sa.Da + Da.Sa + Sa.(1 - Da) + Da.(1 - Sa)
     = Sa + Da
*/

static void
combine_add_8(i_color *out, i_color *in, int channels, i_img_dim count) {
  int ch;
  int color_channels = i_color_channels(channels);
  i_img_dim work_count = count;
  i_color *inp = in;
  i_color *outp = out;

  if (i_has_alpha(channels)) {
    while (work_count--) {
      int src_alpha = inp->channel[color_channels];
      if (src_alpha) {
	int orig_alpha = outp->channel[color_channels];
	int dest_alpha = src_alpha + orig_alpha;
	if (dest_alpha > 255)
	  dest_alpha = 255;
	for (ch = 0; ch < color_channels; ++ch) { 
	  int total = (outp->channel[ch] * orig_alpha + inp->channel[ch] * src_alpha) / dest_alpha;
	  if (total > 255)
	    total = 255;
	  outp->channel[ch] = total;
	}
	outp->channel[color_channels] = dest_alpha;
      }
      
      ++outp;
      ++inp;
    }
  }
  else {
    while (work_count--) {
      int src_alpha = inp->channel[color_channels];
      if (src_alpha) {
	for (ch = 0; ch < color_channels; ++ch) { 
	  int total = outp->channel[ch] + inp->channel[ch] * src_alpha / 255;
	  if (total > 255)
	    total = 255;
	  outp->channel[ch] = total;
	} 
      }
      
      ++outp;
      ++inp;
    }
  }
}

/* 
   [pegtop] documents this as max(A+B-256, 0) while [savvy] documents
   it as max(A-B, 0). [svg1.2] doesn't cover it.

   [savvy] doesn't document how it works with an alpha channel.  GIMP
   actually seems to calculate the final value then use the alpha
   channel to apply that to the target.
 */
static void
combine_subtract_8(i_color *out, i_color *in, int channels, i_img_dim count) {
  int ch;
  i_color const *inp = in;
  i_color *outp = out;
  i_img_dim work_count = count;
  int color_channels = i_color_channels(channels);

  if (i_has_alpha(channels)) {
    while (work_count--) {
      int src_alpha = inp->channel[color_channels];
      if (src_alpha) {
	int orig_alpha = outp->channel[color_channels];
	int dest_alpha = src_alpha + orig_alpha;
	if (dest_alpha > 255)
	  dest_alpha = 255;
	for (ch = 0; ch < color_channels; ++ch) { 
	  int total = 
	    (outp->channel[ch] * orig_alpha - inp->channel[ch] * src_alpha) 
	    / dest_alpha;
	  if (total < 0)
	    total = 0;
	  outp->channel[ch] = total;
	}
	outp->channel[color_channels] = dest_alpha;
      }
      ++outp;
      ++inp;
    }
  }
  else {
    while (work_count--) {
      int src_alpha = inp->channel[color_channels];
      if (src_alpha) {
	for (ch = 0; ch < color_channels; ++ch) { 
	  int total = outp->channel[ch] - inp->channel[ch] * src_alpha / 255;
	  if (total < 0)
	    total = 0;
	  outp->channel[ch] = total;
	} 
      }
      ++outp;
      ++inp;
    }
  }
}

#ifdef IM_EIGHT_BIT
#define IM_abs(x) abs(x)
#else
#define IM_abs(x) fabs(x)
#endif

/*
Dca' = abs(Dca.Sa - Sca.Da) + Sca.(1 - Da) + Dca.(1 - Sa)
     = Sca + Dca - 2.min(Sca.Da, Dca.Sa)
Da'  = Sa + Da - Sa.Da 
*/
static void
combine_diff_8(i_color *out, i_color *in, int channels, i_img_dim count) {
  int ch;
  i_color const *inp = in;
  i_color *outp = out;
  i_img_dim work_count = count;
  int color_channels = i_color_channels(channels);

  if (i_has_alpha(channels)) {
    while (work_count--) {
      int src_alpha = inp->channel[color_channels];
      if (src_alpha) {
	int orig_alpha = outp->channel[color_channels];
	int dest_alpha = src_alpha + orig_alpha 
	  - src_alpha * orig_alpha / 255;
	for (ch = 0; ch < color_channels; ++ch) {
	  int src = inp->channel[ch] * src_alpha;
	  int orig = outp->channel[ch] * orig_alpha;
	  int src_da = src * orig_alpha;
	  int dest_sa = orig * src_alpha;
	  int diff = src_da < dest_sa ? src_da : dest_sa;
	  outp->channel[ch] = (src + orig - 2 * diff / 255) / dest_alpha;
	}
	outp->channel[color_channels] = dest_alpha;
      }
      ++inp;
      ++outp;
    }
  }
  else {
    while (work_count--) {
      int src_alpha = inp->channel[color_channels];
      if (src_alpha) {
	for (ch = 0; ch < color_channels; ++ch) {
	  int src = inp->channel[ch] * src_alpha;
	  int orig = outp->channel[ch] * 255;
	  int src_da = src * 255;
	  int dest_sa = orig * src_alpha;
	  int diff = src_da < dest_sa ? src_da : dest_sa;
	  outp->channel[ch] = (src + orig - 2 * diff / 255) / 255;
	}
      }
      ++inp;
      ++outp;
    }
  }
}

#undef IM_abs

/*
  Dca' = min(Sca.Da, Dca.Sa) + Sca.(1 - Da) + Dca(1 - Sa)
  Da' = Sa + Da - Sa.Da

  To hoist some code:

  Dca' = min(Sc.Sa.Da, Dc.Da.Sa) + Sca - Sca.Da + Dca - Dca.Sa
       = Sa.Da.min(Sc, Dc) + Sca - Sca.Da + Dca - Dca.Sa

  When Da=1:

  Dca' = min(Sca.1, Dc.1.Sa) + Sca.(1 - 1) + Dc.1(1 - Sa)
       = min(Sca, Dc.Sa) + Dc(1-Sa)
       = Sa.min(Sc, Dc) + Dc - Dc.Sa
  Da'  = Sa + 1 - Sa.1
       = 1
 */
static void 
combine_darken_8(i_color *out, i_color *in, int channels, 
			  i_img_dim count) {
  int ch;
  i_color const *inp = in;
  i_color *outp = out;
  i_img_dim work_count = count;
  int color_channels = i_color_channels(channels);

  if (i_has_alpha(channels)) {
    while (work_count--) {
      int src_alpha = inp->channel[color_channels];

      if (src_alpha) {
	int orig_alpha = outp->channel[color_channels];
	int dest_alpha = src_alpha + orig_alpha 
	  - src_alpha * orig_alpha / 255;
	for (ch = 0; ch < color_channels; ++ch) { 
	  int Sca = inp->channel[ch] * src_alpha;
	  int Dca = outp->channel[ch] * orig_alpha;
	  int ScaDa = Sca * orig_alpha;
	  int DcaSa = Dca * src_alpha;
	  int minc = ScaDa < DcaSa ? ScaDa : DcaSa;
	  outp->channel[ch] = 
	    ( 
	     minc + (Sca + Dca) * 255
	     - ScaDa - DcaSa
	     ) / (255 * dest_alpha);
	} 
	outp->channel[color_channels] = dest_alpha;
      }
      ++outp;
      ++inp;
    }
  }
  else {
    while (work_count--) {
      int src_alpha = inp->channel[color_channels];

      if (src_alpha) {
	for (ch = 0; ch < color_channels; ++ch) { 
	  int minc = outp->channel[ch] < inp->channel[ch]
	    ? outp->channel[ch] : inp->channel[ch];
	  outp->channel[ch] = 
	    ( 
	     src_alpha * minc + 
	     outp->channel[ch] * ( 255 - src_alpha )
	     ) / 255;
	} 
      }
      ++outp;
      ++inp;
    }
  }
}

static void 
combine_lighten_8(i_color *out, i_color *in, int channels, i_img_dim count) {
  int ch;
  i_color const *inp = in;
  i_color *outp = out;
  i_img_dim work_count = count;
  int color_channels = i_color_channels(channels);

  if (i_has_alpha(channels)) {
    while (work_count--) {
      int src_alpha = inp->channel[color_channels];

      if (src_alpha) {
	int orig_alpha = outp->channel[color_channels];
	int dest_alpha = src_alpha + orig_alpha 
	  - src_alpha * orig_alpha / 255;
	for (ch = 0; ch < color_channels; ++ch) { 
	  int Sca = inp->channel[ch] * src_alpha;
	  int Dca = outp->channel[ch] * orig_alpha;
	  int ScaDa = Sca * orig_alpha;
	  int DcaSa = Dca * src_alpha;
	  int maxc = ScaDa > DcaSa ? ScaDa : DcaSa;
	  outp->channel[ch] = 
	    ( 
	     maxc + (Sca + Dca) * 255
	     - ScaDa - DcaSa
	     ) / (255 * dest_alpha);
	} 
	outp->channel[color_channels] = dest_alpha;
      }
      ++outp;
      ++inp;
    }
  }
  else {
    while (work_count--) {
      int src_alpha = inp->channel[color_channels];

      if (src_alpha) {
	for (ch = 0; ch < color_channels; ++ch) { 
	  int maxc = outp->channel[ch] > inp->channel[ch]
	    ? outp->channel[ch] : inp->channel[ch];
	  outp->channel[ch] = 
	    ( 
	     src_alpha * maxc + 
	     outp->channel[ch] * ( 255 - src_alpha )
	     ) / 255;
	} 
      }
      ++outp;
      ++inp;
    }
  }
}

#if IM_EIGHT_BIT
#define IM_RGB_TO_HSV i_rgb_to_hsv
#define IM_HSV_TO_RGB i_hsv_to_rgb
#else
#define IM_RGB_TO_HSV i_rgb_to_hsvf
#define IM_HSV_TO_RGB i_hsv_to_rgbf
#endif

static void 
combine_hue_8(i_color *out, i_color *in, int channels, i_img_dim count) {
  if (channels > 2) {
    i_color *inp = in;
    i_color const *outp = out;
    i_img_dim work_count = count;

    if (i_has_alpha(channels)) {
      while (work_count--) {
	i_color c = *inp;
	IM_RGB_TO_HSV(&c);
	/* only transfer hue if there's saturation */
	if (c.channel[1] && inp->channel[3] && outp->channel[3]) {
	  *inp = *outp;
	  IM_RGB_TO_HSV(inp);
	  /* and no point in setting the target hue if the target has no sat */
	  if (inp->channel[1]) {
	    inp->channel[0] = c.channel[0];
	    IM_HSV_TO_RGB(inp);
	    inp->channel[3] = c.channel[3];
	  }
	  else {
	    inp->channel[3] = 0;
	  }
	}
	else {
	  inp->channel[3] = 0;
	}
	
	++outp;
	++inp;
      }
    }
    else {
      while (work_count--) {
	i_color c = *inp;
	IM_RGB_TO_HSV(&c);
	/* only transfer hue if there's saturation */
	if (c.channel[1] && inp->channel[3]) {
	  *inp = *outp;
	  IM_RGB_TO_HSV(inp);
	  /* and no point in setting the target hue if the target has no sat */
	  if (inp->channel[1]) {
	    inp->channel[0] = c.channel[0];
	    IM_HSV_TO_RGB(inp);
	    inp->channel[3] = c.channel[3];
	  }
	}
	else {
	  inp->channel[3] = 0;
	}
	
	++outp;
	++inp;
      }
    }

    combine_line_na_8(out, in, channels, count);
  }
}

static void
combine_sat_8(i_color *out, i_color *in, int channels, i_img_dim count) {
  if (channels > 2) {
    i_color *inp = in;
    i_color const *outp = out;
    i_img_dim work_count = count;

    while (work_count--) {
      i_color c = *inp;
      *inp = *outp;
      IM_RGB_TO_HSV(&c);
      IM_RGB_TO_HSV(inp);
      inp->channel[1] = c.channel[1];
      IM_HSV_TO_RGB(inp);
      inp->channel[3] = c.channel[3];
      ++outp;
      ++inp;
    }

    combine_line_na_8(out, in, channels, count);
  }
}

static void 
combine_value_8(i_color *out, i_color *in, int channels, i_img_dim count) {
  if (channels > 2) {
    i_color *inp = in;
    i_color const *outp = out;
    i_img_dim work_count = count;

    while (work_count--) {
      i_color c = *inp;
      *inp = *outp;
      IM_RGB_TO_HSV(&c);
      IM_RGB_TO_HSV(inp);
      inp->channel[2] = c.channel[2];
      IM_HSV_TO_RGB(inp);
      inp->channel[3] = c.channel[3];
      ++outp;
      ++inp;
    }
  }

  /* all images have a "value channel" - for greyscale it's the only
     colour channel */
  combine_line_na_8(out, in, channels, count);
}

static void 
combine_color_8(i_color *out, i_color *in, int channels, i_img_dim count) {
  if (channels > 2) {
    i_color *inp = in;
    i_color const *outp = out;
    i_img_dim work_count = count;

    while (work_count--) {
      i_color c = *inp;
      *inp = *outp;
      IM_RGB_TO_HSV(&c);
      IM_RGB_TO_HSV(inp);
      inp->channel[0] = c.channel[0];
      inp->channel[1] = c.channel[1];
      IM_HSV_TO_RGB(inp);
      inp->channel[3] = c.channel[3];
      ++outp;
      ++inp;
    }

    combine_line_na_8(out, in, channels, count);
  }
}

#undef IM_RGB_TO_HSV
#undef IM_HSV_TO_RGB

#undef IM_EIGHT_BIT
#undef IM_FILL_COMBINE
#define IM_FILL_COMBINE(fill) ((fill)->combinef)
#undef IM_FILL_FILLER
#define IM_FILL_FILLER(fill) ((fill)->f_fill_with_fcolor)
#line 759 "render.im"

/*
  Three good references for implementing combining modes:

  http://www.w3.org/TR/2004/WD-SVG12-20041027/rendering.html
  referenced as [svg1.2]

  http://gimp-savvy.com/BOOK/index.html?node55.html
  ("The Blending Modes", if it changes)
  referenced as [savvy]

  http://www.pegtop.net/delphi/articles/blendmodes/
  referenced as [pegtop]

  Where differences exist, I follow the SVG practice, the gimp
  practice, and lastly pegtop.
*/


static void 
combine_alphablend_double(i_fcolor *out, i_fcolor *in, int channels, i_img_dim count) {
  combine_line_double(out, in, channels, count);
}

/*
Dca' = Sca.Dca + Sca.(1 - Da) + Dca.(1 - Sa)
Da'  = Sa.Da + Sa.(1 - Da) + Da.(1 - Sa)
     = Sa + Da - Sa.Da

When Da=1

Dc' = Sc.Sa.Dc + Dc.(1 - Sa)
 */
static void
combine_mult_double(i_fcolor *out, i_fcolor *in, int channels, i_img_dim count) {
  int ch;
  i_fcolor *inp = in;
  i_fcolor *outp = out;
  i_img_dim work_count = count;
  int color_channels = i_color_channels(channels);

  if (i_has_alpha(channels)) {
    while (work_count--) {
      double orig_alpha = outp->channel[color_channels];
      double src_alpha = inp->channel[color_channels];
      
      if (src_alpha) {
	double dest_alpha = src_alpha + orig_alpha 
	  - (src_alpha * orig_alpha) / 1.0;
	
	for (ch = 0; ch < color_channels; ++ch) { 
	  outp->channel[ch] = 
	    (inp->channel[ch] * src_alpha * outp->channel[ch] / 1.0
	     * orig_alpha
	     + inp->channel[ch] * src_alpha * (1.0 - orig_alpha)
	     + outp->channel[ch] * orig_alpha * (1.0 - src_alpha))
	    / 1.0 / dest_alpha;
	}
	outp->channel[color_channels] = dest_alpha;
      }
      ++outp;
      ++inp;
    }
  }
  else {
    while (work_count--) {
      double src_alpha = inp->channel[color_channels];
      double remains = 1.0 - src_alpha;
      
      if (src_alpha) {
	for (ch = 0; ch < color_channels; ++ch) { 
	  outp->channel[ch] = 
	    (src_alpha * inp->channel[ch] * outp->channel[ch] / 1.0
	     + outp->channel[ch] * remains) / 1.0;
	}
      }
      ++outp;
      ++inp;
    }
  }
}

static void 
combine_dissolve_double(i_fcolor *out, i_fcolor *in, int channels, i_img_dim count) {
  int color_channels = i_color_channels(channels);
  int ch;

  if (i_has_alpha(channels)) {
    while (count--) {
      if (in->channel[channels-1] > rand() * ((double)1.0 / RAND_MAX)) {
	for (ch = 0; ch < color_channels; ++ch) {
	  out->channel[ch] = in->channel[ch];
	}
	out->channel[color_channels] = 1.0;
      }
      ++out;
      ++in;
    }
  }
  else {
    while (count--) {
      if (in->channel[channels] > rand() * ((double)1.0 / RAND_MAX)) {
	for (ch = 0; ch < color_channels; ++ch) {
	  out->channel[ch] = in->channel[ch];
	}
      }
      ++out;
      ++in;
    }
  }
}

/*
Dca' = Sca.Da + Dca.Sa + Sca.(1 - Da) + Dca.(1 - Sa)
     = Sca + Dca
Da'  = Sa.Da + Da.Sa + Sa.(1 - Da) + Da.(1 - Sa)
     = Sa + Da
*/

static void
combine_add_double(i_fcolor *out, i_fcolor *in, int channels, i_img_dim count) {
  int ch;
  int color_channels = i_color_channels(channels);
  i_img_dim work_count = count;
  i_fcolor *inp = in;
  i_fcolor *outp = out;

  if (i_has_alpha(channels)) {
    while (work_count--) {
      double src_alpha = inp->channel[color_channels];
      if (src_alpha) {
	double orig_alpha = outp->channel[color_channels];
	double dest_alpha = src_alpha + orig_alpha;
	if (dest_alpha > 1.0)
	  dest_alpha = 1.0;
	for (ch = 0; ch < color_channels; ++ch) { 
	  double total = (outp->channel[ch] * orig_alpha + inp->channel[ch] * src_alpha) / dest_alpha;
	  if (total > 1.0)
	    total = 1.0;
	  outp->channel[ch] = total;
	}
	outp->channel[color_channels] = dest_alpha;
      }
      
      ++outp;
      ++inp;
    }
  }
  else {
    while (work_count--) {
      double src_alpha = inp->channel[color_channels];
      if (src_alpha) {
	for (ch = 0; ch < color_channels; ++ch) { 
	  double total = outp->channel[ch] + inp->channel[ch] * src_alpha / 1.0;
	  if (total > 1.0)
	    total = 1.0;
	  outp->channel[ch] = total;
	} 
      }
      
      ++outp;
      ++inp;
    }
  }
}

/* 
   [pegtop] documents this as max(A+B-256, 0) while [savvy] documents
   it as max(A-B, 0). [svg1.2] doesn't cover it.

   [savvy] doesn't document how it works with an alpha channel.  GIMP
   actually seems to calculate the final value then use the alpha
   channel to apply that to the target.
 */
static void
combine_subtract_double(i_fcolor *out, i_fcolor *in, int channels, i_img_dim count) {
  int ch;
  i_fcolor const *inp = in;
  i_fcolor *outp = out;
  i_img_dim work_count = count;
  int color_channels = i_color_channels(channels);

  if (i_has_alpha(channels)) {
    while (work_count--) {
      double src_alpha = inp->channel[color_channels];
      if (src_alpha) {
	double orig_alpha = outp->channel[color_channels];
	double dest_alpha = src_alpha + orig_alpha;
	if (dest_alpha > 1.0)
	  dest_alpha = 1.0;
	for (ch = 0; ch < color_channels; ++ch) { 
	  double total = 
	    (outp->channel[ch] * orig_alpha - inp->channel[ch] * src_alpha) 
	    / dest_alpha;
	  if (total < 0)
	    total = 0;
	  outp->channel[ch] = total;
	}
	outp->channel[color_channels] = dest_alpha;
      }
      ++outp;
      ++inp;
    }
  }
  else {
    while (work_count--) {
      double src_alpha = inp->channel[color_channels];
      if (src_alpha) {
	for (ch = 0; ch < color_channels; ++ch) { 
	  double total = outp->channel[ch] - inp->channel[ch] * src_alpha / 1.0;
	  if (total < 0)
	    total = 0;
	  outp->channel[ch] = total;
	} 
      }
      ++outp;
      ++inp;
    }
  }
}

#ifdef IM_EIGHT_BIT
#define IM_abs(x) abs(x)
#else
#define IM_abs(x) fabs(x)
#endif

/*
Dca' = abs(Dca.Sa - Sca.Da) + Sca.(1 - Da) + Dca.(1 - Sa)
     = Sca + Dca - 2.min(Sca.Da, Dca.Sa)
Da'  = Sa + Da - Sa.Da 
*/
static void
combine_diff_double(i_fcolor *out, i_fcolor *in, int channels, i_img_dim count) {
  int ch;
  i_fcolor const *inp = in;
  i_fcolor *outp = out;
  i_img_dim work_count = count;
  int color_channels = i_color_channels(channels);

  if (i_has_alpha(channels)) {
    while (work_count--) {
      double src_alpha = inp->channel[color_channels];
      if (src_alpha) {
	double orig_alpha = outp->channel[color_channels];
	double dest_alpha = src_alpha + orig_alpha 
	  - src_alpha * orig_alpha / 1.0;
	for (ch = 0; ch < color_channels; ++ch) {
	  double src = inp->channel[ch] * src_alpha;
	  double orig = outp->channel[ch] * orig_alpha;
	  double src_da = src * orig_alpha;
	  double dest_sa = orig * src_alpha;
	  double diff = src_da < dest_sa ? src_da : dest_sa;
	  outp->channel[ch] = (src + orig - 2 * diff / 1.0) / dest_alpha;
	}
	outp->channel[color_channels] = dest_alpha;
      }
      ++inp;
      ++outp;
    }
  }
  else {
    while (work_count--) {
      double src_alpha = inp->channel[color_channels];
      if (src_alpha) {
	for (ch = 0; ch < color_channels; ++ch) {
	  double src = inp->channel[ch] * src_alpha;
	  double orig = outp->channel[ch] * 1.0;
	  double src_da = src * 1.0;
	  double dest_sa = orig * src_alpha;
	  double diff = src_da < dest_sa ? src_da : dest_sa;
	  outp->channel[ch] = (src + orig - 2 * diff / 1.0) / 1.0;
	}
      }
      ++inp;
      ++outp;
    }
  }
}

#undef IM_abs

/*
  Dca' = min(Sca.Da, Dca.Sa) + Sca.(1 - Da) + Dca(1 - Sa)
  Da' = Sa + Da - Sa.Da

  To hoist some code:

  Dca' = min(Sc.Sa.Da, Dc.Da.Sa) + Sca - Sca.Da + Dca - Dca.Sa
       = Sa.Da.min(Sc, Dc) + Sca - Sca.Da + Dca - Dca.Sa

  When Da=1:

  Dca' = min(Sca.1, Dc.1.Sa) + Sca.(1 - 1) + Dc.1(1 - Sa)
       = min(Sca, Dc.Sa) + Dc(1-Sa)
       = Sa.min(Sc, Dc) + Dc - Dc.Sa
  Da'  = Sa + 1 - Sa.1
       = 1
 */
static void 
combine_darken_double(i_fcolor *out, i_fcolor *in, int channels, 
			  i_img_dim count) {
  int ch;
  i_fcolor const *inp = in;
  i_fcolor *outp = out;
  i_img_dim work_count = count;
  int color_channels = i_color_channels(channels);

  if (i_has_alpha(channels)) {
    while (work_count--) {
      double src_alpha = inp->channel[color_channels];

      if (src_alpha) {
	double orig_alpha = outp->channel[color_channels];
	double dest_alpha = src_alpha + orig_alpha 
	  - src_alpha * orig_alpha / 1.0;
	for (ch = 0; ch < color_channels; ++ch) { 
	  double Sca = inp->channel[ch] * src_alpha;
	  double Dca = outp->channel[ch] * orig_alpha;
	  double ScaDa = Sca * orig_alpha;
	  double DcaSa = Dca * src_alpha;
	  double minc = ScaDa < DcaSa ? ScaDa : DcaSa;
	  outp->channel[ch] = 
	    ( 
	     minc + (Sca + Dca) * 1.0
	     - ScaDa - DcaSa
	     ) / (1.0 * dest_alpha);
	} 
	outp->channel[color_channels] = dest_alpha;
      }
      ++outp;
      ++inp;
    }
  }
  else {
    while (work_count--) {
      double src_alpha = inp->channel[color_channels];

      if (src_alpha) {
	for (ch = 0; ch < color_channels; ++ch) { 
	  double minc = outp->channel[ch] < inp->channel[ch]
	    ? outp->channel[ch] : inp->channel[ch];
	  outp->channel[ch] = 
	    ( 
	     src_alpha * minc + 
	     outp->channel[ch] * ( 1.0 - src_alpha )
	     ) / 1.0;
	} 
      }
      ++outp;
      ++inp;
    }
  }
}

static void 
combine_lighten_double(i_fcolor *out, i_fcolor *in, int channels, i_img_dim count) {
  int ch;
  i_fcolor const *inp = in;
  i_fcolor *outp = out;
  i_img_dim work_count = count;
  int color_channels = i_color_channels(channels);

  if (i_has_alpha(channels)) {
    while (work_count--) {
      double src_alpha = inp->channel[color_channels];

      if (src_alpha) {
	double orig_alpha = outp->channel[color_channels];
	double dest_alpha = src_alpha + orig_alpha 
	  - src_alpha * orig_alpha / 1.0;
	for (ch = 0; ch < color_channels; ++ch) { 
	  double Sca = inp->channel[ch] * src_alpha;
	  double Dca = outp->channel[ch] * orig_alpha;
	  double ScaDa = Sca * orig_alpha;
	  double DcaSa = Dca * src_alpha;
	  double maxc = ScaDa > DcaSa ? ScaDa : DcaSa;
	  outp->channel[ch] = 
	    ( 
	     maxc + (Sca + Dca) * 1.0
	     - ScaDa - DcaSa
	     ) / (1.0 * dest_alpha);
	} 
	outp->channel[color_channels] = dest_alpha;
      }
      ++outp;
      ++inp;
    }
  }
  else {
    while (work_count--) {
      double src_alpha = inp->channel[color_channels];

      if (src_alpha) {
	for (ch = 0; ch < color_channels; ++ch) { 
	  double maxc = outp->channel[ch] > inp->channel[ch]
	    ? outp->channel[ch] : inp->channel[ch];
	  outp->channel[ch] = 
	    ( 
	     src_alpha * maxc + 
	     outp->channel[ch] * ( 1.0 - src_alpha )
	     ) / 1.0;
	} 
      }
      ++outp;
      ++inp;
    }
  }
}

#if IM_EIGHT_BIT
#define IM_RGB_TO_HSV i_rgb_to_hsv
#define IM_HSV_TO_RGB i_hsv_to_rgb
#else
#define IM_RGB_TO_HSV i_rgb_to_hsvf
#define IM_HSV_TO_RGB i_hsv_to_rgbf
#endif

static void 
combine_hue_double(i_fcolor *out, i_fcolor *in, int channels, i_img_dim count) {
  if (channels > 2) {
    i_fcolor *inp = in;
    i_fcolor const *outp = out;
    i_img_dim work_count = count;

    if (i_has_alpha(channels)) {
      while (work_count--) {
	i_fcolor c = *inp;
	IM_RGB_TO_HSV(&c);
	/* only transfer hue if there's saturation */
	if (c.channel[1] && inp->channel[3] && outp->channel[3]) {
	  *inp = *outp;
	  IM_RGB_TO_HSV(inp);
	  /* and no point in setting the target hue if the target has no sat */
	  if (inp->channel[1]) {
	    inp->channel[0] = c.channel[0];
	    IM_HSV_TO_RGB(inp);
	    inp->channel[3] = c.channel[3];
	  }
	  else {
	    inp->channel[3] = 0;
	  }
	}
	else {
	  inp->channel[3] = 0;
	}
	
	++outp;
	++inp;
      }
    }
    else {
      while (work_count--) {
	i_fcolor c = *inp;
	IM_RGB_TO_HSV(&c);
	/* only transfer hue if there's saturation */
	if (c.channel[1] && inp->channel[3]) {
	  *inp = *outp;
	  IM_RGB_TO_HSV(inp);
	  /* and no point in setting the target hue if the target has no sat */
	  if (inp->channel[1]) {
	    inp->channel[0] = c.channel[0];
	    IM_HSV_TO_RGB(inp);
	    inp->channel[3] = c.channel[3];
	  }
	}
	else {
	  inp->channel[3] = 0;
	}
	
	++outp;
	++inp;
      }
    }

    combine_line_na_double(out, in, channels, count);
  }
}

static void
combine_sat_double(i_fcolor *out, i_fcolor *in, int channels, i_img_dim count) {
  if (channels > 2) {
    i_fcolor *inp = in;
    i_fcolor const *outp = out;
    i_img_dim work_count = count;

    while (work_count--) {
      i_fcolor c = *inp;
      *inp = *outp;
      IM_RGB_TO_HSV(&c);
      IM_RGB_TO_HSV(inp);
      inp->channel[1] = c.channel[1];
      IM_HSV_TO_RGB(inp);
      inp->channel[3] = c.channel[3];
      ++outp;
      ++inp;
    }

    combine_line_na_double(out, in, channels, count);
  }
}

static void 
combine_value_double(i_fcolor *out, i_fcolor *in, int channels, i_img_dim count) {
  if (channels > 2) {
    i_fcolor *inp = in;
    i_fcolor const *outp = out;
    i_img_dim work_count = count;

    while (work_count--) {
      i_fcolor c = *inp;
      *inp = *outp;
      IM_RGB_TO_HSV(&c);
      IM_RGB_TO_HSV(inp);
      inp->channel[2] = c.channel[2];
      IM_HSV_TO_RGB(inp);
      inp->channel[3] = c.channel[3];
      ++outp;
      ++inp;
    }
  }

  /* all images have a "value channel" - for greyscale it's the only
     colour channel */
  combine_line_na_double(out, in, channels, count);
}

static void 
combine_color_double(i_fcolor *out, i_fcolor *in, int channels, i_img_dim count) {
  if (channels > 2) {
    i_fcolor *inp = in;
    i_fcolor const *outp = out;
    i_img_dim work_count = count;

    while (work_count--) {
      i_fcolor c = *inp;
      *inp = *outp;
      IM_RGB_TO_HSV(&c);
      IM_RGB_TO_HSV(inp);
      inp->channel[0] = c.channel[0];
      inp->channel[1] = c.channel[1];
      IM_HSV_TO_RGB(inp);
      inp->channel[3] = c.channel[3];
      ++outp;
      ++inp;
    }

    combine_line_na_double(out, in, channels, count);
  }
}

#undef IM_RGB_TO_HSV
#undef IM_HSV_TO_RGB

#line 1314 "render.im"
